(* +------------------------------------------------------------------------+
   |                            Arbres de priorit                          |
   +------------------------------------------------------------------------+ *)

load_object "arbres.zo";;
#open "arbres";;
install_printer "print_int_b_arbre";;

(* compte le nombre de comparaisons *)
let cmp = ref 0;;

let lt = prefix <;;
let le = prefix <=;;
let gt = prefix >;;
let ge = prefix >=;;

let prefix <  a b = incr cmp; lt a b;;
let prefix <= a b = incr cmp; le a b;;
let prefix >  a b = incr cmp; gt a b;;
let prefix >= a b = incr cmp; ge a b;;

(* insertion *)
let rec insertion x a = match a with
  | B_vide         -> B_noeud(x,B_vide,B_vide)
  | B_noeud(y,g,d) -> if y > x then B_noeud(x, d, insertion y g)
                               else B_noeud(y, d, insertion x g)
;;

(* runion de deux arbres et d'une racine *)
let rec union x g d = match g,d with
  | B_noeud(y,gy,dy), B_noeud(z,gz,dz) ->
           if (x > y) & (z >  y) then B_noeud(y, union x gy dy, d)
      else if (x > z) & (z <= y) then B_noeud(z, g, union x gz dz)
      else B_noeud(x,g,d)
  | B_vide, B_noeud(z,gz,dz) ->
      if x > z then B_noeud(z, g, union x gz dz)
      else B_noeud(x,g,d)
  | _,_ -> B_noeud(x,g,d)
;;

(* extraction de l'lment le plus  droite *)
let rec extrait_droite a = match a with
  | B_vide              -> failwith "arbre vide"
  | B_noeud(x,_,B_vide) -> (x,B_vide)
  | B_noeud(x,g,d)      -> let (y,d') = extrait_droite d in (y, B_noeud(x,d',g))
;;

(* extraction de la racine, mthode 1 *)
let extrait_racine_1 a = match a with
  | B_vide              -> failwith "arbre vide"
  | B_noeud(x,_,B_vide) -> (x,B_vide)
  | B_noeud(x,g,d)      -> let (y,d') = extrait_droite d in (x, union y d' g)
;;

(* extraction de la racine, mthode 2 *)
let rec extrait_racine_2 a = match a with
  | B_vide              -> failwith "arbre vide"
  | B_noeud(x,B_vide,d) -> (x, d)
  | B_noeud(x,B_noeud(y,gy,dy),d) ->
      let z,d' = extrait_racine_2 d in
      if y >= z then (x, B_noeud(z,d',B_noeud(y,gy,dy)))
                else (x, B_noeud(y, d', union z gy dy))
;;

(* arbre alatoire *)
let rec random_ape n a b =
  if n = 0 then B_vide
  else let x = a + random__int(b-a+1)
       and y = random_ape (n-1) a b in insertion x y
;;

(* vrification : retourne un boolen disant si l'arbre est un APE 
                           et sa taille *)
let rec verif a = match a with
  | B_vide -> true,0
  | B_noeud(x,g,d) -> 
      let bong,tg = verif g in
      let bond,td = verif d in
      let bon =
	bong & bond & (tg <= td) & (td <= tg + 1)
        & (match g with B_vide -> true | B_noeud(y,_,_) -> y >= x)
        & (match d with B_vide -> true | B_noeud(y,_,_) -> y >= x)
      in (bon, tg+td+1)
;;


(* comparaison des deux mthodes d'extraction *)
let moyenne n p =
  let c1 = ref 0 and c2 = ref 0 in
  for i=1 to p do
    let a = random_ape n 0 10000 in
    cmp := 0;
    let _ = extrait_racine_1 a in
    c1 := !c1 + !cmp;
    cmp := 0;
    let _ = extrait_racine_2 a in
    c2 := !c2 + !cmp
  done;
  (float_of_int(!c1)/.float_of_int(p),
   float_of_int(!c2)/.float_of_int(p))
;;

(* rsultats :

moyenne 10 100;;
- : float * float = 6.08, 3.41
#moyenne 100 100;;
- : float * float = 16.43, 12.88
#moyenne 1000 100;;
- : float * float = 27.22, 22.4
#moyenne 10000 100;;
- : float * float = 39.3, 31.23
#

Donc la mthode 2 fait environ 20% de comparaisons que la mthode 1.
De plus, le nb moyen de comparaisons augmente de 12 environ (10 pour M2)
lorsque n est multipli par 10, on a bien une complexit logarithmique.
*)

(* suite de Hamming *)
let hamming(n) =
  let a = ref(B_noeud(1,B_vide,B_vide)) in
  let l = ref []     in
  let compte = ref 0 in
  while !compte < n do
    let x,b = extrait_racine_2 !a in
    match !l with
      | y::_ when y = x -> a := b
      | _ -> l := x :: !l;
             a := insertion (2*x) b;
             a := insertion (3*x) !a;
             a := insertion (5*x) !a;
             compte := !compte + 1
  done;
  rev !l
;;


(* amlioration *)
let hamming_1 n =
  let a = ref(B_noeud(1,B_vide,B_vide)) in
  let l = ref [] in
  for i=1 to n do
    let x,b = extrait_racine_2 !a in
    l := x :: !l;
    a := insertion (5*x) b;
    if x mod 5 > 0 then begin
      a := insertion (3*x) !a;
      if x mod 3 > 0 then a := insertion (2*x) !a
    end
  done;
  rev !l
;;


         (* +--------------------------------------------+
            |  Calcul rapide du n-me entier de Hamming  |
            +--------------------------------------------+ *)

(* transformation Ak -> A(k+1) *)
let transforme (a,impairs,coef,taille) =
  let b = ref(a) and imp = ref(impairs) and c = 2*coef and t = ref(taille) in
  while
    match !imp with
    | B_noeud(x,_,_) -> x < c
    | _ -> failwith "cas impossible"
  do
    let x,i = extrait_racine_2 !imp in
    imp := insertion (5*x) i;
    if x mod 5 > 0 then imp := insertion (3*x) !imp;
    b := insertion (float_of_int(x)/.float_of_int(c)) !b;
    t := !t + 1
  done;
  (!b,!imp,c,!t)
;;

(* n-me entier de Hamming *)
let hamming_rapide(n) =
  let tat = ref(B_vide,                   (* a       *)
                 B_noeud(1,B_vide,B_vide), (* impairs *)
                 1,                        (* coef    *)
                 0)                        (* taille  *)
  and compte = ref(0) in

  (* acumule les entiers de Hamming par tranches entre deux puissances de 2 *)
  while !compte <= n do
    tat := transforme !tat;
    let (_,_,_,t) = !tat in compte := !compte + t
  done;

  (* extrait le n-(compte prcdent) me *)
  let (a,_,c,t) = !tat in
  let b = ref a in
  for i = !compte-t to n-1 do let _,b' = extrait_racine_2 !b in b := b' done;
  match !b with B_noeud(x,_,_) -> int_of_float(float_of_int(c)*.x)
              | B_vide -> failwith "impossible"
;;

(* rsultat :
#cmp := 0; hd(rev(hamming 101));; !cmp;;
- : int = 1600
#- : int = 3936
#cmp := 0; hamming_rapide 100;; !cmp;;
- : int = 1600
#- : int = 324
#cmp := 0; hd(rev(hamming 1001));; !cmp;;
- : int = 51840000
#- : int = 92209
#cmp := 0; hamming_rapide 1000;; !cmp;;
- : int = 51840000
#- : int = 2524
#
*)
